package n4l;

import static java.util.logging.Level.WARNING;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.nio.channels.DatagramChannel;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.ResourceBundle;
import java.util.logging.Logger;

import org.apache.oltu.oauth2.common.exception.OAuthProblemException;
import org.apache.oltu.oauth2.common.exception.OAuthSystemException;

import n4l.model.Module;
import n4l.model.Station;
import n4l.netatmo.NetatmoHttpClient;


public class Netatmo4Loxone implements Runnable {

	private static final String		VERSION		= "1.0";

	// ======================================================================================
	// -------- configuration properties are stored in Configuration.properties file --------

	// Miniserver specific data - CHANGE IT TO YOUR NEEDS!
	static private String			miniserverAddress;
	static private int				miniserverPort;

	// Netatmo App data - CHANGE IT TO YOUR APP DATA!
	static private String			CLIENT_ID;
	static private String			CLIENT_SECRET;

	// Netatmo Account data - CHANGE IT TO YOUR USER DATA!
	static private String			EMAIL;
	static private String			PASSWORD;

	// set this to true/false to enable/disable continuous trace of UDP Messages
	static private boolean			traceMsUdp;
	
	// set this to true/false to enable/disable continuous debug logging	
	private static boolean			debugNetatmoHttp;
	private static boolean			debugDataExtractor;

	// set this to true if you have more than one Netatmo Station and like to distinguish them easily
	// if set to true, qualified module names will be used: <StationName>.<ModuleName>.<Key>
	// example: Burggasse.Esszimmer.CO2
	// if set to false, simple module names will be used: <ModuleName>.<Key>
	// example: Esszimmer.CO2
	static private boolean			useQualifiedNames;

	// -------- configuration properties are stored in Configuration.properties file --------
	// ======================================================================================

	private static final DateFormat	formatter	= new SimpleDateFormat("yyyy.MM.dd HH:mm:ss");
	private boolean					firstRun	= true;
	private static final Logger		log			= Logger.getLogger(Netatmo4Loxone.class.getName());


	/**
	 * One Main Class to to rule them all
	 * 
	 * @param args
	 */
	public static void main(String args[]) {

		// set this to INFO to enable log messages
		log.setLevel(WARNING);

		readRessourceBundle();

		// start the forever running thread
		(new Thread(new Netatmo4Loxone())).start();
	}


	private static void readRessourceBundle() {
		ResourceBundle rb = ResourceBundle.getBundle("Configuration");

		miniserverAddress = rb.getString("MINISERVER_ADDRESS");
		miniserverPort = Integer.parseInt(rb.getString("MINISERVER_PORT"));

		CLIENT_ID = rb.getString("NETATMO_CLIENT_ID");
		CLIENT_SECRET = rb.getString("NETATMO_CLIENT_SECRET");

		EMAIL = rb.getString("NETAMO_USER_EMAIL");
		PASSWORD = rb.getString("NETATMO_USER_PASSWORD");

		traceMsUdp = Boolean.parseBoolean(rb.getString("TRACE_MS_UDP_DATA"));
		debugNetatmoHttp = Boolean.parseBoolean(rb.getString("DEBUG_NETATMO_ACCESS"));
		debugDataExtractor = Boolean.parseBoolean(rb.getString("DEBUG_DATA_EXTRACTION"));

		useQualifiedNames = Boolean.parseBoolean(rb.getString("USE_QUALIFIED_NAMES"));
	}


	@Override
	public void run() {

		System.out.println("Hello Miniserver, this is Netatmo4Loxone version " + VERSION + " calling you at address "
				+ miniserverAddress + ":" + miniserverPort);

		NetatmoHttpClient client = new NetatmoHttpClient(CLIENT_ID, CLIENT_SECRET);

		while (true) {
			try {
				client.login(EMAIL, PASSWORD);

				List<Station> stations = client.getStationData();

				if (firstRun) {
					showStations(stations);
				}

				if (firstRun || traceMsUdp) {
					System.out.println("\nUDP Data sent to Miniserver " + miniserverAddress + ":" + miniserverPort
							+ " at " + formatter.format(new Date()));
				}

				for (Station station : stations) {
					for (Module module : station.getModules()) {
						String kvValues = module.getAllValuesAsKVString(" ", useQualifiedNames);
						sendUDPData(kvValues);
						if (firstRun || traceMsUdp) {
							System.out.println(kvValues);
						}
						// Miniserver seems to loose some udp data if telegrams come to fast, I could see them in the
						// UDP Monitor but sometimes a value update was missing, so there is a short delay after each telegram
						sleep(113);
					}
				}

				firstRun = false;

				// Netatmo publishes new data every 10 minutes, so we check for new data every 4 minutes
				// see Nyquist–Shannon sampling theorem: fSampl > 2 * fMax
				sleep(1000 * 60 * 1);
			}
			catch (OAuthProblemException oape) {
				log.severe(() -> oape.getError() + " - Execution stopped, check your Netatmo account configuration settings!");
				oape.printStackTrace();
				System.exit(-1);
			}
			catch (OAuthSystemException oase) {
				if (oase.getCause() instanceof UnknownHostException) {
					log.warning(() -> oase.getCause().getMessage() + " - Can not connect to host, check your network settings! Will retry in 30 seconds.");
					firstRun = true;
					sleep(1000 * 30);
				}
				else {
					log.severe(() -> oase.getMessage() + " - Unhandled error, check Stack trace! Will retry in 30 seconds.");
					oase.printStackTrace();					
					firstRun = true;
					sleep(1000 * 30);
				}
			}
		}
	}


	private void sleep(int milis) {
		try {
			Thread.sleep(milis);
		}
		catch (InterruptedException e1) {
		}
	}


	/**
	 * Show all the Stations
	 * 
	 * @param stations
	 */
	private void showStations(List<Station> stations) {
		System.out.println("\nThis are your Stations and their Modules");
		for (Station s : stations) {
			System.out.println(s);
		}
	}


	/**
	 * Send UDP data to defined Miniserver address:port
	 * 
	 * @param data
	 */
	public static void sendUDPData(String data) {
		try (DatagramChannel channel = DatagramChannel.open()) {

			ByteBuffer buf = ByteBuffer.allocate(3000);
			buf.put(data.getBytes());
			buf.flip();

			InetSocketAddress socket = new InetSocketAddress(miniserverAddress, miniserverPort);
			channel.send(buf, socket);
		}
		catch (IOException ioe) {
			log.severe(() -> ioe.getMessage() + " - Could not send UDP Data to Miniserver!");
			ioe.printStackTrace();
		}
	}


	public static boolean getDebugNetatmoHTTPCalss() {
		return debugNetatmoHttp;
	}
	
	
	public static boolean getDebugDataExtractorf() {
		return debugDataExtractor;
	}

}
