/*
 * Copyright 2015 Maximilian Weißböck, Wien
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package n4l.netatmo;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import n4l.Netatmo4Loxone;
import n4l.model.Module;
import n4l.model.Station;


public class NetatmoDataExtractor {
			
	private static final Logger log = Logger.getLogger(NetatmoDataExtractor.class.getName());
	

	/**
	 * Extract all data from all stations
	 * 
	 * @param response
	 * @return
	 */
	public static List<Station> parseDevicesListAndExtractData(JSONObject response) {
		
		log.setLevel(Level.WARNING);
		if (Netatmo4Loxone.getDebugDataExtractorf()) {
			log.setLevel(Level.INFO);			
		}

		List<Station> stations = new ArrayList<Station>();

		try {
			JSONArray jsonStations = response.getJSONObject("body").getJSONArray("devices");
			log.info(() -> "The complete JSON  object with all the Stations, Modules and Data\n" + jsonStations.toString());

			// parse all the stations data and build all Station objects with their Base Module
			for (int i = 0; i < jsonStations.length(); i++) {
				JSONObject jsonStation = jsonStations.getJSONObject(i);
				String name = jsonStation.getString("station_name");
				String moduleName = jsonStation.getString("module_name");
				String id = jsonStation.getString("_id");
				String type = jsonStation.getString("type");

				Station station = new Station(name, id);
				Module mainModule = new Module(station, moduleName, id, type);

				station.addModule(mainModule);
				stations.add(station);

				extractModuleData(mainModule, jsonStation);
				log.info(() -> "The extractet Data as key=value String from the main Module\n" 
							+ mainModule.getAllValuesAsKVString("\n", false));

				JSONArray jsonModuels = jsonStation.getJSONArray("modules");
				log.info(() -> "The JSON  object witha all the Modules and their Data\n" + jsonModuels.toString());

				// parse all the Modules (excluding Base Module), build them and add them to their Station
				for (int j = 0; j < jsonModuels.length(); j++) {
					JSONObject jsonModule = jsonModuels.getJSONObject(j);
					log.info(() -> "The JSON object for one Module and its Data\n" + jsonModule.toString());

					moduleName = jsonModule.getString("module_name");
					id = jsonModule.getString("_id");
					type = jsonModule.getString("type");

					Module module = new Module(station, moduleName, id, type);
					station.addModule(module);

					extractModuleData(module, jsonModule);
					log.info(() -> "The extractet Data as key=value String from one Module\n" 
							+ module.getAllValuesAsKVString("\n", false));
				}
			}
		}
		catch (JSONException e) {
			e.printStackTrace();
		}

		return stations;
	}


	/**
	 * Extract the data for a Netatmo Module form the JASON message and store in with the Module Object
	 * 
	 * @param module
	 * @param jsonModule
	 */
	private static void extractModuleData(Module module, JSONObject jsonModule) {

//		
//		Complete JASON Example Data 
//		
//		[{"station_name":"Burggasse","cipher_id":"enc:16:LnRWRUl16A/h1b6hUOVEeCR/xcKeVaouhJ6QIzNxRACx8kaobBkMnvSdeXCjpHDr","type":"NAMain","invitation_disable":false,
//
//			"modules":[{"rf_status":75,"battery_vp":5380,"last_seen":1449566536,"battery_percent":74,
//							"data_type":["Temperature","Humidity"],
//							"last_message":1449566555,"_id":"02:00:00:00:42:70","module_name":"Außen",
//							"dashboard_data":{"date_min_temp":1449529319,"time_utc":1449566536,"date_max_temp":1449566228,"Temperature":2.9,"min_temp":2,"temp_trend":"up","Humidity":91,"max_temp":2.9},
//							"type":"NAModule1","firmware":43},
//						{"rf_status":64,"battery_vp":5523,"last_seen":1449566536,"battery_percent":74,
//							"data_type":["Temperature","CO2","Humidity"],
//							"last_message":1449566555,"_id":"03:00:00:00:6f:9a","module_name":"Schlafzimmer",
//							"dashboard_data":{"date_min_temp":1449554745,"time_utc":1449566536,"date_max_temp":1449566228,"Temperature":20.4,"min_temp":17.6,"temp_trend":"stable",
//								"Humidity":47,"CO2":905,"max_temp":20.4},
//							"type":"NAModule4","firmware":43},
//						{"rf_status":72,"battery_vp":5681,"last_seen":1449566536,"battery_percent":82,
//							"data_type":["Temperature","CO2","Humidity"],
//							"last_message":1449566555,"_id":"03:00:00:00:74:32","module_name":"Kinderzimmer",
//							"dashboard_data":{"date_min_temp":1449554437,"time_utc":1449566536,"date_max_temp":1449566228,"Temperature":21.6,"min_temp":19.5,"temp_trend":"stable",
//								"Humidity":43,"CO2":1006,"max_temp":21.6},
//							"type":"NAModule4","firmware":43}
//		
//						{"rf_status":120,"battery_vp":6062,"last_seen":1451842265,"battery_percent":100,
//							"data_type":["Rain"],
//							"last_message":1451842271,"_id":"05:00:00:01:f4:56","module_name":"Regenmesser",
//							"dashboard_data":{"time_utc":1451842252,"Rain":0,"sum_rain_24":0.202,"sum_rain_1":0},
//							"type":"NAModule3","firmware":8},
//							
//						{"rf_status":74,"battery_vp":6302,"last_seen":1451842271,"battery_percent":100,
//							"data_type":["Wind"],
//							"last_message":1451842271,"_id":"06:00:00:00:1d:06","module_name":"Windmesser",
//							"dashboard_data":{"date_max_wind_str":1451776040,"date_min_temp":1451775739,"time_utc":1451842258,
//								"WindHistoric":[
//									{"time_utc":1451838946,"WindAngle":101,"WindStrength":7},
//									{"time_utc":1451839247,"WindAngle":90,"WindStrength":5},
//									{"time_utc":1451839547,"WindAngle":103,"WindStrength":7},
//									{"time_utc":1451839848,"WindAngle":102,"WindStrength":6},
//									{"time_utc":1451840150,"WindAngle":108,"WindStrength":7},
//									{"time_utc":1451840451,"WindAngle":90,"WindStrength":5},
//									{"time_utc":1451840752,"WindAngle":98,"WindStrength":4},
//									{"time_utc":1451841053,"WindAngle":101,"WindStrength":7},
//									{"time_utc":1451841355,"WindAngle":90,"WindStrength":4},
//									{"time_utc":1451841656,"WindAngle":96,"WindStrength":6},
//									{"time_utc":1451841957,"WindAngle":90,"WindStrength":6},
//									{"time_utc":1451842258,"WindAngle":97,"WindStrength":6}],
//								"GustAngle":107,"max_wind_str":28,"date_max_temp":1451775739,"max_wind_angle":78,"min_temp":0,
//								"GustStrength":11,"max_temp":0,
//								"WindAngle":97,
//								"WindStrength":6},
//							"type":"NAModule2","firmware":11}
//						],
//		
//			"last_upgrade":1440263707,"co2_calibrating":false,"co2_abc_status":"co2_abc","read_only":true,"wifi_status":53,
//			"data_type":["Temperature","CO2","Humidity","Noise","Pressure"],
//			"_id":"70:ee:50:00:42:aa","module_name":"Esszimmer",
//			"place":{"altitude":202,"country":"AT","city":"","timezone":"Europe/Vienna","geoip_city":"Vienna","improveLocProposed":true,"location":[16.342585611768,48.204536142579]},
//			"dashboard_data":{"date_min_temp":1449555661,"time_utc":1449566544,"Temperature":22.3,"pressure_trend":"stable","Noise":57,"temp_trend":"stable","CO2":1107,
//				"Pressure":1036.7,"date_max_temp":1449565938,"min_temp":19.4,"Humidity":40,"AbsolutePressure":1012.2,"max_temp":22.3},
//			"firmware":111,"last_status_store":1449566558}]
//

		// get Module data
		if (module.getType().equals(Module.MAIN_STATION)) { // Data from Main Module
			setLongValue(module, "wifi_status", jsonModule);
		}
		else { // Data from other Modules
			setLongValue(module, "battery_percent", jsonModule);
			setLongValue(module, "rf_status", jsonModule);
			setLongValue(module, "battery_vp", jsonModule);
			setTimeValue(module, "last_seen", jsonModule);
			setTimeValue(module, "last_message", jsonModule);
		}

		JSONObject jsonDashboardData = jsonModule.getJSONObject("dashboard_data");
		setTimeValue(module, "time_utc", jsonDashboardData);

		// what special data is inside dashboard (Temperature, Humidity, CO2, Noise, Rain, Wind) is defined by data_type
		JSONArray jsonDataTypes = jsonModule.getJSONArray("data_type");
		for (int i = 0; i < jsonDataTypes.length(); i++) {
			String dataType = jsonDataTypes.getString(i);
			log.info(() -> "dataType:" + dataType);
			if (dataType.equals("Temperature")) { // all Temperature Values
				setDoubleValue(module, dataType, jsonDashboardData);
				setDoubleValue(module, "max_temp", jsonDashboardData);				
				setTimeValue(module, "date_max_temp", jsonDashboardData);
				setDoubleValue(module, "min_temp", jsonDashboardData);
				setTimeValue(module, "date_min_temp", jsonDashboardData);
				setStringValue(module, "temp_trend", jsonDashboardData);
			}
			else if (dataType.equals("Rain")) {
				setDoubleValue(module, dataType, jsonDashboardData);
				setDoubleValue(module, "sum_rain_24", jsonDashboardData);
				setDoubleValue(module, "sum_rain_1", jsonDashboardData);
			}
			else if (dataType.equals("Wind")) {
				setTimeValue(module, "date_max_wind_str", jsonDashboardData);
				setLongValue(module, "max_wind_angle", jsonDashboardData);
				setLongValue(module, "max_wind_str", jsonDashboardData);
				setLongValue(module, "WindAngle", jsonDashboardData);
				setLongValue(module, "WindStrength", jsonDashboardData);
				setLongValue(module, "GustAngle", jsonDashboardData);
				setLongValue(module, "GustStrength", jsonDashboardData);				
			}
			else if (dataType.equals("Pressure")) {
				setLongValue(module, dataType, jsonDashboardData);			
				setLongValue(module, "AbsolutePressure", jsonDashboardData);				
				setStringValue(module, "pressure_trend", jsonDashboardData);
			}
			else { // Humidity, CO2, Noise
				setLongValue(module, dataType, jsonDashboardData);
			}
		}
	}


	/**
	 * set a String Object
	 * 
	 * @param module
	 * @param key
	 * @param jsonObject
	 */
	private static void setStringValue(Module module, String key, JSONObject jsonObject) {
		if (jsonObject.has(key)) {
			module.setValue(key, jsonObject.getString(key));
		}
		else {
			module.setValue(key, key);
			log.warning(() -> "No value found for " + module.getQualifiedName() + "." + key + " in jsonObject:" + jsonObject.toString()
					+ "\nWill set value '" + key + "' instead - this will trigger another warning if this value is read");
		}
	}


	/**
	 * set a Long Object
	 * 
	 * @param module
	 * @param key
	 * @param jsonObject
	 */
	private static void setLongValue(Module module, String key, JSONObject jsonObject) {
		if (jsonObject.has(key)) {
			module.setValue(key, jsonObject.getLong(key));
		}
		else {
			module.setValue(key, -999L);
			log.warning(() -> "No value found for " + module.getQualifiedName() + "." + key + " in jsonObject:" + jsonObject.toString()
					+ "\nWill set value -999 insted - this shall be interpreted as an illegal value");
		}
	}


	/**
	 * set a double Object
	 * 
	 * @param module
	 * @param key
	 * @param jsonObject
	 */
	private static void setDoubleValue(Module module, String key, JSONObject jsonObject) {
		if (jsonObject.has(key)) {
			module.setValue(key, jsonObject.getDouble(key));
		}
		else {
			module.setValue(key, -999D);
			log.warning(() -> "No value found for " + module.getQualifiedName() + "." + key + " in jsonObject:" + jsonObject.toString()
			+ "\nWill set value -999 insted - this shall be interpreted as an illegal value");
			log.warning(() -> "No value found for " + module.getQualifiedName() + "." + key + " in jsonObject:" + jsonObject.toString()
			+ "\nWill set value -999 insted - this shall be interpreted as an illegal value");
		}
	}


	// seconds between 1.1.1970 (base for UTC Time) and 1.1.2009 (base for Miniserver Time)
	private static final long seconds1970to2009 = 1230768000L;
	
	/**
	 * set a Time Value
	 * 
	 * Miniserver expects time values in seconds since 1.1.2009
	 * This is in contrast to the normal time which is based on 1.1.1970
	 * So we need to adapt this
	 * 
	 * @param module
	 * @param key
	 * @param jsonObject
	 */
	private static void setTimeValue(Module module, String key, JSONObject jsonObject) {
		if (jsonObject.has(key)) {			
			
			long time = ((Long)jsonObject.getLong(key)).longValue();
			
			// time from Netatmo is in UTC
			Instant instant = Instant.ofEpochSecond(time);
			if (key.equals("time_utc")) {
				module.setLastUpdate(time);
			}

			// convert to system time
			LocalDateTime ldt = LocalDateTime.ofInstant(instant, ZoneId.systemDefault());
			log.info(() -> "Local Time:" + ldt.toString());
			
			// we need this to get the seconds
			ZonedDateTime zdt = ldt.atZone(ZoneId.systemDefault());
			long offset = zdt.getOffset().getTotalSeconds(); 
			log.info(() -> "Local Time Offset to UTC in seconds:" + offset);
			
			// convert to Miniserver time			
			// epoch seconds is based on UTC, so we need the offset and the diff from 1970 to 2009
			long localDate = zdt.toEpochSecond() + offset - seconds1970to2009;
			// this gets the local time, so we need no offset
			long localTime = zdt.getHour() * 3600 + zdt.getMinute() * 60 + zdt.getSecond();
									
			module.setValue(key, new Long(localDate));
			module.setValue(key + "_tod", new Long(localTime));
			// Miniserver Status Baustein does not handle time value <v1.t> as expected, so we send a number value 
			// use <v1> to display it looks a little strange -  11.12 instead 11:12 - but there seems no other solution so far
			double timeAsNmbr = zdt.getHour() + zdt.getMinute() / 100.0;			
			module.setValue(key + "_tod_nmbr", timeAsNmbr);
		}
		else {
			module.setValue(key, 0L);
			log.warning(() -> "key:" + key + " not found in jsonObject:\n" + jsonObject.toString());
		}
	}
	
}